/*
===========================================================================
Copyright (C) 1999-2005 Id Software, Inc.
Copyright (C) 2005 Stuart Dalton (badcdev@gmail.com)

This file is part of Quake III Arena source code.

Quake III Arena source code is free software; you can redistribute it
and/or modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of the License,
or (at your option) any later version.

Quake III Arena source code is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Foobar; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
===========================================================================
*/

/*
 * snd_al_buffers.c
 * Buffer allocation and management for the OpenAL sound system
 *
 * 2005-08-22
 *  Started logging
 *  Split buffer management out from snd_al.c
 *
 * 2005-08-23
 *  Reworked to use sfxHandles instead of re-cast pointers. What was I thinking?
 *  Added error checking
 *  Started on sound unloader
 */

#include "../client/client.h"
#include "snd_al_local.h"

typedef struct sfx_s sfx_t;
struct sfx_s
{
	char 		filename[MAX_QPATH];
	ALuint		buffer;		// OpenAL buffer
	qboolean	isDefault;	// Couldn't be loaded - use default FX
	qboolean	inMemory;	// Sound is stored in memory
	qboolean	isLocked;	// Sound is locked (can not be unloaded)
	int		used;		// Time last used
	sfx_t 		*next;		// Next entry in hash list
};

static qboolean al_buffer_inited = qfalse;

/**
 * Sound effect storage, data structures
 */
#define MAX_SFX 4096
sfx_t knownSfx[MAX_SFX];
int numSfx;

sfxHandle_t default_sfx;

// Find a free handle
sfxHandle_t al_buf_find_free()
{
	int i;

	for(i = 0; i < MAX_SFX; i++)
	{
		// Got one
		if(knownSfx[i].filename[0] == '\0')
			return i;
	}

	// Shit...
	Com_Error(ERR_FATAL, "al_buf_find_free: No free sound handles");
}

// Find a sound effect if loaded, set up a handle otherwise
sfxHandle_t al_buf_find(const char *filename)
{
	// Look it up in the hash table
	sfxHandle_t sfx = -1;
	int i;

	for(i = 0; i < MAX_SFX; i++)
	{
		if(!Q_stricmp(knownSfx[i].filename, filename))
		{
			sfx = i;
			break;
		}
	}

	// Not found in hash table?
	if(sfx == -1)
	{
		sfx_t *ptr;
		long hash;

		sfx = al_buf_find_free();

		// Clear and copy the filename over
		ptr = &knownSfx[sfx];
		Com_Memset(ptr, 0, sizeof(*ptr));
		strcpy(ptr->filename, filename);
	}

	// Return the handle
	return sfx;
}

/**
 * Initialisation and shutdown
 */
// Called at init, shutdown
qboolean al_buf_init()
{
	sfxHandle_t sfx;

	if(al_buffer_inited)
		return qtrue;

	// Clear the hash table, and SFX table
	Com_Memset(knownSfx, 0, sizeof(knownSfx));

	// Load the default sound, and lock it
	default_sfx = al_buf_find("sound/feedback/hit.wav");
	al_buf_use(default_sfx);
	knownSfx[default_sfx].isLocked = qtrue;

	// All done
	al_buffer_inited = qtrue;
}

void al_buf_shutdown()
{
	int i;

	if(!al_buffer_inited)
		return;

	// Unlock the default sound effect
	knownSfx[default_sfx].isLocked = qfalse;

	// Free all used effects
	for(i = 0; i < MAX_SFX; i++)
		al_buf_unload(i);

	// Clear the tables
	Com_Memset(knownSfx, 0, sizeof(knownSfx));

	// All undone
	al_buffer_inited = qfalse;
}

/**
 * Registration
 */
sfxHandle_t al_buf_register(const char *filename)
{
	sfxHandle_t sfx = al_buf_find(filename);

	if((s_precache->integer == 1) && (!knownSfx[sfx].inMemory) && (!knownSfx[sfx].isDefault))
		al_buf_load(sfx);
	knownSfx[sfx].used = Com_Milliseconds();

	return sfx;
}

/**
 * Usage counter
 */
void al_buf_use(sfxHandle_t sfx)
{
	if(knownSfx[sfx].filename[0] == '\0')
		return;

	if((!knownSfx[sfx].inMemory) && (!knownSfx[sfx].isDefault))
		al_buf_load(sfx);
	knownSfx[sfx].used = Com_Milliseconds();
}

/**
 * Loading and unloading
 */
static void al_buf_use_default(sfxHandle_t sfx)
{
	if(sfx == default_sfx)
		Com_Error(ERR_FATAL, "Can't load default sound effect %s\n", knownSfx[sfx].filename);

	Com_Printf("Warning: Using default sound for %s\n", knownSfx[sfx].filename);
	knownSfx[sfx].isDefault = qtrue;
	knownSfx[sfx].buffer = knownSfx[default_sfx].buffer;
}

void al_buf_load(sfxHandle_t sfx)
{
	byte *src_data;

	ALenum error;
	ALenum format;
	ALvoid *data;
	ALsizei size;
	ALsizei freq;

#ifndef MACINTOSH_AL
	ALboolean loop;
#endif

	// Nothing?
	if(knownSfx[sfx].filename[0] == '\0')
		return;

	// Player SFX
	if(knownSfx[sfx].filename[0] == '*')
		return;

	// Already done?
	if((knownSfx[sfx].inMemory) || (knownSfx[sfx].isDefault))
		return;

	// Read the file into memory
	FS_ReadFile(knownSfx[sfx].filename, (void **)&src_data);
	if(!src_data)
	{
		// Use the default sound effect
		al_buf_use_default(sfx);
		return;
	}

	// Extract data
	alGetError();
#ifndef MACINTOSH_AL
	alutLoadWAVMemory(src_data, &format, &data, &size, &freq, &loop);
#else
	alutLoadWAVMemory(src_data, &format, &data, &size, &freq);
#endif
	FS_FreeFile(src_data);

	if((error = alGetError()) != AL_NO_ERROR)
	{
		Com_Printf("alutLoadWAVMemory failed\n");
		al_buf_use_default(sfx);
		return;
	}

	// Create a buffer
	alGenBuffers(1, &knownSfx[sfx].buffer);
	if((error = alGetError()) != AL_NO_ERROR)
		Com_Error(ERR_FATAL, "Can't create a sound buffer for %s", knownSfx[sfx].filename);

	// Fill the buffer
	alBufferData(knownSfx[sfx].buffer, format, data, size, freq);
	error = alGetError();

	// If we ran out of memory, start evicting the least recently used sounds
	while(error == AL_OUT_OF_MEMORY)
	{
		qboolean rv = al_buf_evict();
		if(!rv)
			Com_Error(ERR_FATAL, "Catastrophic sound memory exhaustion!");

		// Try load it again
		alBufferData(knownSfx[sfx].buffer, format, data, size, freq);
		error = alGetError();
	}

	// Some other error condition
	if(error != AL_NO_ERROR)
		Com_Error(ERR_FATAL, "Can't fill sound buffer for %s", knownSfx[sfx].filename);

	// Free the memory OpenAL allocated for us
	alutUnloadWAV(format, data, size, freq);

	// Woo!
	knownSfx[sfx].inMemory = qtrue;
}

void al_buf_unload(sfxHandle_t sfx)
{
	ALenum error;

	if(knownSfx[sfx].filename[0] == '\0')
		return;

	if(!knownSfx[sfx].inMemory)
		return;

	// Delete it
	alDeleteBuffers(1, &knownSfx[sfx].buffer);
	if((error = alGetError()) != AL_NO_ERROR)
		Com_Printf("Can't delete sound buffer for %s", knownSfx[sfx].filename);

	knownSfx[sfx].inMemory = qfalse;
}

qboolean al_buf_evict()
{
	// Doesn't work yet, so if OpenAL reports that you're out of memory, you'll just get
	// "Catastrophic sound memory exhaustion". Whoops.
	return qfalse;
}

/**
 * Buffer grabbage
 */
ALuint al_buf_get(sfxHandle_t sfx)
{
	return knownSfx[sfx].buffer;
}

char *al_buf_get_name(sfxHandle_t sfx)
{
	return knownSfx[sfx].filename;
}

