/*
===========================================================================
Copyright (C) 1999-2005 Id Software, Inc.
Copyright (C) 2005 Stuart Dalton (badcdev@gmail.com)

This file is part of Quake III Arena source code.

Quake III Arena source code is free software; you can redistribute it
and/or modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of the License,
or (at your option) any later version.

Quake III Arena source code is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Foobar; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
===========================================================================
*/

/*
 * snd_al_main.c
 * Main section of OpenAL audio system
 *
 * 2005-08-22
 *  Started logging
 *  Split buffer management out from snd_al.c
 *
 * 2005-08-23
 *  Reworked to use sfxHandles instead of re-cast pointers. What was I thinking?
 *  New initialization code using alc instead of alut
 *  Renamed to snd_al_main.c
 *  Split source management out from snd_al_main.c, reworked it a bit
 *  More error checking
 *  Moved music stuff to snd_al_music.c
 */

#include "../client/client.h"
#include "snd_al_local.h"

// I honestly don't know what these are supposed to be for...
int s_soundtime = 0;
int s_rawend = 0;

void S_Play_f( void );
void S_Music_f( void );

/**
 * OpenAL error messages
 */
char *al_errormsg(ALenum error)
{
	switch(error)
	{
		case AL_NO_ERROR:
			return "No error";
		case AL_INVALID_NAME:
			return "Invalid name";
		case AL_INVALID_ENUM:
			return "Invalid enumerator";
		case AL_INVALID_VALUE:
			return "Invalid value";
		case AL_INVALID_OPERATION:
			return "Invalid operation";
		case AL_OUT_OF_MEMORY:
			return "Out of memory";
		default:
			return "Unknown error";
	}
}

char *alc_errormsg(ALenum error)
{
	switch(error)
	{
		case ALC_NO_ERROR:
			return "No error";
		case ALC_INVALID_DEVICE:
			return "Invalid device";
		case ALC_INVALID_CONTEXT:
			return "Invalid context";
		case ALC_INVALID_ENUM:
			return "Invalid enumerator";
		case ALC_INVALID_VALUE:
			return "Invalid value";
		case ALC_OUT_OF_MEMORY:
			return "Out of memory";
		default:
			return "Unknown error";
	}
}

/**
 * Local state variables
 */
qboolean snd_inited = qfalse;
static ALCdevice *alDevice;
static ALCcontext *alContext;

/**
 * Console variables
 */
cvar_t *s_volume;
cvar_t *s_musicVolume;
cvar_t *s_doppler;
cvar_t *s_khz;
cvar_t *s_precache;
cvar_t *s_gain;
cvar_t *s_sources;
cvar_t *s_dopplerFactor;
cvar_t *s_dopplerSpeed;
cvar_t *s_minDistance;
cvar_t *s_rolloff;


void S_Init( void )
{
	ALenum error;
	cvar_t *cv;
	int hertz;
	ALint cxt_attributes [] =
	{
		ALC_FREQUENCY, 13370,
		0, 0
	};

	if(snd_inited)
		return;

	// Original console variables
	s_volume = Cvar_Get ("s_volume", "0.8", CVAR_ARCHIVE);
	s_musicVolume = Cvar_Get ("s_musicvolume", "0.25", CVAR_ARCHIVE);
	s_doppler = Cvar_Get ("s_doppler", "1", CVAR_ARCHIVE);
	s_khz = Cvar_Get ("s_khz", "44", CVAR_ARCHIVE);

	// New console variables
	s_precache = Cvar_Get ("s_precache", "0", CVAR_ARCHIVE);
	s_gain = Cvar_Get ("s_gain", "0.25", CVAR_ARCHIVE);
	s_sources = Cvar_Get ("s_sources", "64", CVAR_ARCHIVE);
	s_dopplerFactor = Cvar_Get ("s_dopplerfactor", "1.0", CVAR_ARCHIVE);
	s_dopplerSpeed = Cvar_Get ("s_dopplerspeed", "2200", CVAR_ARCHIVE);
	s_minDistance = Cvar_Get ("s_mindistance", "80", CVAR_ARCHIVE);
	s_rolloff = Cvar_Get ("s_rolloff", "0.25", CVAR_ARCHIVE);

	Com_Printf("\n------- sound initialization -------\n");

	cv = Cvar_Get ("s_initsound", "1", 0);
	if ( !cv->integer ) {
		Com_Printf ("not initializing.\n");
		Com_Printf("------------------------------------\n");
		return;
	}

	// Initialize the OpenAL default device
	Com_Printf("using device %s\n", alcGetString(NULL, ALC_DEFAULT_DEVICE_SPECIFIER));
	alDevice = alcOpenDevice(NULL);
	if((error = alcGetError(NULL)) != ALC_NO_ERROR)
	{
		Com_Printf("can not initialize device: %s (0x%04X)\n", alc_errormsg(error), error);
		Com_Printf("------------------------------------\n");
		return;
	}

	// Work out sampling rate in khz
	switch(s_khz->integer)
	{
		case 11:
			hertz = 11025;
			break;
		case 22:
			hertz = 22050;
			break;
		case 44:
			hertz = 44100;
			break;
		default:
			hertz = s_khz->integer * 1000;
			break;
	}

	// Set up attributes
	Com_Printf("attr(1) = %d\n", cxt_attributes[1]);
	cxt_attributes[1] = hertz;
	Com_Printf("using %dHz sampling rate\n", hertz);

	// Try to get an OpenAL context
	alContext = alcCreateContext(alDevice, cxt_attributes);
	if((error = alcGetError(NULL)) != ALC_NO_ERROR)
	{
		Com_Printf("can not create context: %s (0x%04X)\n", alc_errormsg(error), error);
		Com_Printf("------------------------------------\n");
		return;
	}

	// Activate context
	alcMakeContextCurrent(alContext);
	if((error = alcGetError(NULL)) != ALC_NO_ERROR)
	{
		Com_Printf("can not activate context %s (0x%04X)\n", alc_errormsg(error), error);
		Com_Printf("------------------------------------\n");
		return;
	}

	// Initialize sources, buffers, music
	al_buf_init();
	al_src_init();
	al_mus_init();

	// Set up OpenAL parameters (doppler, etc)
	alDopplerFactor(s_dopplerFactor->value);
	alDopplerVelocity(s_dopplerSpeed->value);

	// Add commands
	Cmd_AddCommand("play", S_Play_f);
	Cmd_AddCommand("music", S_Music_f);

	// Init successful
	Com_Printf("initialization successful\n");
	Com_Printf("------------------------------------\n");
	snd_inited = qtrue;
}

void S_Shutdown( void )
{
	if(!snd_inited)
		return;

	// Shut down buffers, sources
	al_mus_shutdown();
	al_src_shutdown();
	al_buf_shutdown();

	// Kill the OpenAL context, device
	// For some reason, this makes Q3 hang on shutdown
//	alcMakeContextCurrent(NULL);
	alcDestroyContext(alContext);
	alcCloseDevice(alDevice);

	snd_inited = qfalse;
}

/**
 * Stub functions
 */
void S_StartSound( vec3_t origin, int entnum, int entchannel, sfxHandle_t sfx )
{
	if(!snd_inited)
		return;

	al_src_play(sfx, origin, entnum, entchannel);
}

void S_StartLocalSound( sfxHandle_t sfx, int channelNum )
{
	if(!snd_inited)
		return;

	al_src_local(sfx, channelNum);
}

// cinematics and voice-over-network will send raw samples
// 1.0 volume will be direct output of source samples
void S_RawSamples (int samples, int rate, int width, int channels, 
		   const byte *data, float volume)
{
	if(!snd_inited)
		return;
}

// stop all sounds and the background track
void S_StopAllSounds( void )
{
	if(!snd_inited)
		return;
	al_src_shutup();
}

// all continuous looping sounds must be added before calling S_Update
void S_ClearLoopingSounds( qboolean killall )
{
	if(!snd_inited)
		return;

	al_src_loop_clear();
}

void S_AddLoopingSound( int entityNum, const vec3_t origin, const vec3_t velocity, sfxHandle_t sfx )
{
	if(!snd_inited)
		return;

	al_src_loop(SRCPRI_AMBIENT, sfx, origin, velocity, entityNum);
}

void S_AddRealLoopingSound( int entityNum, const vec3_t origin, const vec3_t velocity, sfxHandle_t sfx )
{
	if(!snd_inited)
		return;

	al_src_loop(SRCPRI_ENTITY, sfx, origin, velocity, entityNum);
}

void S_StopLoopingSound(int entityNum )
{
	if(!snd_inited)
		return;

	al_src_loop_stop(entityNum);
}

// recompute the reletive volumes for all running sounds
// reletive to the given entityNum / orientation
void S_Respatialize( int entityNum, const vec3_t origin, vec3_t axis[3], int inwater )
{
	// Axis[0] = Forward
	// Axis[2] = Up
	float velocity[] = {0.0f, 0.0f, 0.0f};
	float orientation[] = {axis[0][0], axis[0][1], axis[0][2],
		axis[2][0], axis[2][1], axis[2][2]};

	if(!snd_inited)
		return;

	// Set OpenAL listener paramaters
	alListenerfv(AL_POSITION, origin);
	alListenerfv(AL_VELOCITY, velocity);
	alListenerfv(AL_ORIENTATION, orientation);
}

// let the sound system know where an entity currently is
void S_UpdateEntityPosition( int entityNum, const vec3_t origin )
{
	if(!snd_inited)
		return;

	al_src_entity_move(entityNum, origin);
}

void S_Update( void )
{
	if(!snd_inited)
		return;

	al_src_update();
	al_mus_update();
}

void S_DisableSounds( void )
{
	if(!snd_inited)
		return;

	
}

void S_BeginRegistration( void )
{
	if(!snd_inited)
		return;
}

// RegisterSound will allways return a valid sample, even if it
// has to create a placeholder.  This prevents continuous filesystem
// checks for missing files
sfxHandle_t S_RegisterSound(const char *sample, qboolean compressed)
{
	if(!snd_inited)
		return;

	return al_buf_register(sample);
}

void S_DisplayFreeMemory(void)
{
	if(!snd_inited)
		return;
}

void S_ClearSoundBuffer( void )
{
	if(!snd_inited)
		return;
}

void SNDDMA_Activate( void )
{
	if(!snd_inited)
		return;
}

/**
 * Commands
 */
void S_Play_f( void ) {
	int 		i;
	sfxHandle_t	h;
	char		name[256];
	
	i = 1;
	while ( i<Cmd_Argc() ) {
		if ( !Q_strrchr(Cmd_Argv(i), '.') ) {
			Com_sprintf( name, sizeof(name), "%s.wav", Cmd_Argv(1) );
		} else {
			Q_strncpyz( name, Cmd_Argv(i), sizeof(name) );
		}

		h = S_RegisterSound( name, qfalse );
		Com_Printf("Playing SFX %d\n", h);
		if( h ) {
			S_StartLocalSound( h, CHAN_LOCAL_SOUND );
		}
		i++;
	}
}

void S_Music_f( void ) {
	int		c;

	c = Cmd_Argc();

	if ( c == 2 ) {
		S_StartBackgroundTrack( Cmd_Argv(1), Cmd_Argv(1) );
//		s_backgroundLoop[0] = 0;
	} else if ( c == 3 ) {
		S_StartBackgroundTrack( Cmd_Argv(1), Cmd_Argv(2) );
	} else {
		Com_Printf ("music <musicfile> [loopfile]\n");
		return;
	}

}
