/*
===========================================================================
Copyright (C) 1999-2005 Id Software, Inc.
Copyright (C) 2005 Stuart Dalton (badcdev@gmail.com)

This file is part of Quake III Arena source code.

Quake III Arena source code is free software; you can redistribute it
and/or modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of the License,
or (at your option) any later version.

Quake III Arena source code is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Foobar; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
===========================================================================
*/

/*
 * snd_al_music.c
 * Music playback for the OpenAL sound system
 *
 * 2005-08-23
 *  Started implementing music system
 *
 * 2005-08-24
 *  Finished implementing music system
 *  Added Ogg Vorbis support
 */

#include "../client/client.h"
#include "snd_al_local.h"

#include <vorbis/codec.h>
#include <vorbis/vorbisenc.h>
#include <vorbis/vorbisfile.h>

#define BUFFERS	4
#define	BUFFER_SIZE 4096

static qboolean mus_inited = qfalse;
static qboolean mus_playing = qfalse;
static qboolean mus_vorbis = qfalse;
static srcHandle_t source_handle = -1;
static ALuint source;
static ALuint buffers[BUFFERS];

/**
 * Wave file reading
 */
int FGetLittleLong( fileHandle_t f ) {
	int		v;

	FS_Read( &v, sizeof(v), f );

	return LittleLong( v);
}

int FGetLittleShort( fileHandle_t f ) {
	short	v;

	FS_Read( &v, sizeof(v), f );

	return LittleShort( v);
}

// returns the length of the data in the chunk, or 0 if not found
static int S_FindWavChunk( fileHandle_t f, char *chunk ) {
	char	name[5];
	int		len;
	int		r;

	name[4] = 0;
	len = 0;
	r = FS_Read( name, 4, f );
	if ( r != 4 ) {
		return 0;
	}
	len = FGetLittleLong( f );
	if ( len < 0 || len > 0xfffffff ) {
		len = 0;
		return 0;
	}
	len = (len + 1 ) & ~1;		// pad to word boundary

	if ( strcmp( name, chunk ) ) {
		return 0;
	}

	return len;
}

static void S_ByteSwapRawSamples( int samples, int width, int s_channels, const byte *data ) {
	int		i;

	if ( width != 2 ) {
		return;
	}
	if ( LittleShort( 256 ) == 256 ) {
		return;
	}

	if ( s_channels == 2 ) {
		samples <<= 1;
	}
	for ( i = 0 ; i < samples ; i++ ) {
		((short *)data)[i] = LittleShort( ((short *)data)[i] );
	}
}

typedef struct wavstream_s
{
	fileHandle_t handle;
	int seek_pos;

	// Stream total size, remaining
	int left_bytes;
	int total_bytes;

	// Stuff for S_ByteSwapRawSamples
	int width;		// Sample width in bytes
	int channels;		// Number of sound channels (1 or 2)
	int blocksamples;	// Number of samples per block

	// Read buffer
	byte read_buffer[BUFFER_SIZE];

	// OpenAL format descriptors
	ALenum format;
	ALsizei freq;
} wavstream_t;

static wavstream_t *wavstream_load(const char *filename)
{
	char name[MAX_QPATH];
	char dump[16];
	int wav_format;
	int len;
	int fmt_fail = 0;

	wavstream_t *stream = (wavstream_t *)calloc(1, sizeof(wavstream_t));
	if(!stream)
		return NULL;

	// Tack an extension onto it
	Q_strncpyz( name, filename, sizeof( name ) - 4 );
	COM_DefaultExtension( name, sizeof( name ), ".wav" );

	// Open the wave file
	FS_FOpenFileRead( name, &stream->handle, qtrue );
	if(!stream->handle)
	{
		Com_Printf( S_COLOR_YELLOW "WARNING: couldn't open music file %s\n", name );
		return NULL;
	}

	// skip the riff wav header
	FS_Read(dump, 12, stream->handle);

	// Scan for the format chunk
	if ( !S_FindWavChunk( stream->handle, "fmt " ) ) {
		Com_Printf( "No fmt chunk in %s\n", name );
		FS_FCloseFile( stream->handle );
		free(stream);
		return NULL;
	}

	// Save the paramaters
	wav_format = FGetLittleShort(stream->handle);
	stream->channels = FGetLittleShort(stream->handle);
	stream->freq = FGetLittleLong(stream->handle);
	FGetLittleLong(stream->handle);
	FGetLittleShort(stream->handle);
	stream->width = FGetLittleShort(stream->handle) / 8;
	stream->blocksamples = (BUFFER_SIZE / stream->width) / stream->channels;

	// Has to be PCM (format 1)
	if(wav_format != 1) {
		Com_Printf( "Not a PCM wave file %s\n", name );
		FS_FCloseFile( stream->handle );
		free(stream);
		return NULL;
	}

	// Find the data chunk
	if ( ( len = S_FindWavChunk( stream->handle, "data" ) ) == 0 ) {
		Com_Printf("No data chunk in %s\n", name);
		FS_FCloseFile( stream->handle );
		free(stream);
		return NULL;
	}

	// Length of stream
	stream->left_bytes = stream->total_bytes = len;

	// Start of stream
	stream->seek_pos = FS_FTell(stream->handle);

	// Work out OpenAL format
	if(stream->width == 1)
	{
		if(stream->channels == 1)
			stream->format = AL_FORMAT_MONO8;
		else if(stream->channels == 2)
			stream->format = AL_FORMAT_STEREO8;
		else
			fmt_fail = 1;
	}
	else if(stream->width == 2)
	{
		if(stream->channels == 1)
			stream->format = AL_FORMAT_MONO16;
		else if(stream->channels == 2)
			stream->format = AL_FORMAT_STEREO16;
		else
			fmt_fail = 1;
	}
	else
		fmt_fail = 1;

	if(fmt_fail)
	{
		Com_Printf("Can't determine AL audio format in %s\n", name);
		FS_FCloseFile( stream->handle );
		free(stream);
		return NULL;
	}

	return stream;
}

static void wavstream_kill(wavstream_t *stream)
{
	if(stream)
	{
		FS_FCloseFile( stream->handle );
		free(stream);
	}
}

// Returns the number of bytes read
// If it returns < BUFFER_SIZE, it hit the end and you should finish up
static qboolean wavstream_readbytes(wavstream_t *stream)
{
	int len = BUFFER_SIZE;

	if(len > stream->left_bytes)
		len = stream->left_bytes;
	stream->left_bytes -= len;

	FS_Read(stream->read_buffer, len, stream->handle);
	return len;
}

// Rewind to the beginning of the stream if necessary
static void wavstream_rewind(wavstream_t *stream)
{
	if(stream->left_bytes > 0)
		return;

	FS_Seek(stream->handle, stream->seek_pos, FS_SEEK_SET);
	stream->left_bytes = stream->total_bytes;
}

static wavstream_t *wav_intro_stream;
static wavstream_t *wav_loop_stream;
static wavstream_t *wav_current_stream;

/**
 * Wave streaming
 */
static qboolean StartWaveTrack(const char *intro, const char *loop)
{
	qboolean mus_loop_only = qfalse;

	if(!mus_inited)
		return;

	if((!intro) || (!intro[0]))
	{
		mus_loop_only = qtrue;
		if((!loop) || (!loop[0]))
			return qfalse;
	}
	else if((!loop) || (!loop[0]))
	{
		mus_loop_only = qtrue;
		loop = intro;
	}
	else if(!strcmp(intro, loop))
	{
		mus_loop_only = qtrue;
	}
	else
	{
		mus_loop_only = qfalse;
	}

	// Open the streams
	if(!mus_loop_only)
	{
		wav_intro_stream = wavstream_load(intro);
		if(!wav_intro_stream)
			return qfalse;
	}
	wav_loop_stream = wavstream_load(loop);
	if(!wav_loop_stream)
	{
		wavstream_kill(wav_intro_stream);
		return qfalse;
	}

	// Set the current stream
	if(!mus_loop_only)
		wav_current_stream = wav_intro_stream;
	else
		wav_current_stream = wav_loop_stream;

	mus_vorbis = qfalse;
	return qtrue;
}

static void StopWaveTrack()
{
	if(wav_intro_stream)
		wavstream_kill(wav_intro_stream);
	if(wav_loop_stream)
		wavstream_kill(wav_loop_stream);
	wav_intro_stream = wav_loop_stream = wav_current_stream = NULL;
}

static void UpdateWaveTrack(ALint buf)
{
	int len;
	len = wavstream_readbytes(wav_current_stream);
	
	alBufferData(buf, wav_current_stream->format, wav_current_stream->read_buffer, len, wav_current_stream->freq);

	if(len < BUFFER_SIZE)
	{
		wav_current_stream = wav_loop_stream;
		wavstream_rewind(wav_current_stream);
	}
}

/**
 * Ogg Vorbis filesystem hooks
 */
static fileHandle_t ogg_intro_file;
static fileHandle_t ogg_loop_file;
static fileHandle_t ogg_current_file;
static OggVorbis_File ogg_stream;
static ALuint ogg_format;
static int ogg_freq;

size_t oggcallback_read(void *ptr, size_t size, size_t nmemb, void *datasource)
{
	int rv = FS_Read(ptr, (size * nmemb), ogg_current_file);
	return (rv / size);
}

int oggcallback_seek(void *datasource, ogg_int64_t offset, int whence)
{
	return -1;
}

int oggcallback_close(void *datasource)
{
	// We'll handle that in the streamer
	return 0;
}

long oggcallback_tell(void *datasource)
{
	return FS_FTell(ogg_current_file);
}

ov_callbacks oggcallback_struct;

/**
 * Ogg Vorbis streamer
 */

static qboolean OggOpenStream()
{
	vorbis_info *ogg_info;

	// Open the Vorbis stream on the file
	int open_result = ov_open_callbacks((void *)0xDECAFBAD, &ogg_stream, NULL, 0, oggcallback_struct);
	if(open_result != 0)
	{
		Com_Printf(S_COLOR_YELLOW "WARNING: no Ogg Vorbis bitstream found\n");
		return qfalse;
	}

	// Rip some info out of the file
	ogg_info = ov_info(&ogg_stream, -1);
	if(ogg_info->channels == 1)
		ogg_format = AL_FORMAT_MONO16;
	else
		ogg_format = AL_FORMAT_STEREO16;
	ogg_freq = ogg_info->rate;

	return qtrue;
}

static void OggCloseStream()
{
	ov_clear(&ogg_stream);
}

static qboolean StartVorbisTrack(const char *intro, const char *loop)
{
	char name[MAX_QPATH];
	qboolean mus_loop_only = qfalse;

	if(!mus_inited)
		return;

	if((!intro) || (!intro[0]))
	{
		mus_loop_only = qtrue;
		if((!loop) || (!loop[0]))
			return qfalse;
	}
	else if((!loop) || (!loop[0]))
	{
		mus_loop_only = qtrue;
		loop = intro;
	}
	else if(!strcmp(intro, loop))
	{
		mus_loop_only = qtrue;
	}
	else
	{
		mus_loop_only = qfalse;
	}

	// Open the intro file
	if(!mus_loop_only)
	{
		Q_strncpyz( name, intro, sizeof( name ) - 4 );
		COM_DefaultExtension( name, sizeof( name ), ".ogg" );
		FS_FOpenFileRead( name, &ogg_intro_file, qtrue );
		if(!ogg_intro_file)
		{
			Com_Printf( S_COLOR_YELLOW "WARNING: couldn't open music file %s\n", name );
			return qfalse;
		}
	}

	// Open the loop file
	Q_strncpyz( name, loop, sizeof( name ) - 4 );
	COM_DefaultExtension( name, sizeof( name ), ".ogg" );
	FS_FOpenFileRead( name, &ogg_loop_file, qtrue );
	if(!ogg_loop_file)
	{
		FS_FCloseFile(ogg_intro_file);
		Com_Printf( S_COLOR_YELLOW "WARNING: couldn't open music file %s\n", name );
		return qfalse;
	}

	// Choose the first file
	// Set the current stream
	if(!mus_loop_only)
		ogg_current_file = ogg_intro_file;
	else
		ogg_current_file = ogg_loop_file;

	if(!OggOpenStream())
	{
		if(ogg_intro_file)
			FS_FCloseFile(ogg_intro_file);
		if(ogg_loop_file)
			FS_FCloseFile(ogg_loop_file);
		ogg_intro_file = ogg_loop_file = ogg_current_file = 0;
		return qfalse;
	}

	mus_vorbis = qtrue;
	return qtrue;
}

static void KillVorbisTrack()
{
	OggCloseStream();

	if(ogg_intro_file)
		FS_FCloseFile(ogg_intro_file);
	if(ogg_loop_file)
		FS_FCloseFile(ogg_loop_file);
	ogg_intro_file = ogg_loop_file = ogg_current_file = 0;
}

static void UpdateVorbisTrack(ALint buf)
{
	int size = 0;
	int section;
	byte data[BUFFER_SIZE];

	while(size < BUFFER_SIZE)
	{
		int result = ov_read(&ogg_stream, (data+size), (BUFFER_SIZE-size), 0, 2, 1, &section);
		if(result > 0)
			size += result;
		else if(result == 0)
		{
			Com_Printf("Rewinding...\n");
			OggCloseStream();
			ogg_current_file = ogg_loop_file;
			OggOpenStream();
		}
	}

	alBufferData(buf, ogg_format, data, size, ogg_freq);
	/*
	int len;
	len = wavstream_readbytes(wav_current_stream);
	
	alBufferData(buf, wav_current_stream->format, wav_current_stream->read_buffer, len, wav_current_stream->freq);

	if(len < BUFFER_SIZE)
	{
	wav_current_stream = wav_loop_stream;
	wavstream_rewind(wav_current_stream);
}*/

	/*
	char data[BUFFER_SIZE];
	int  size = 0;
	int  section;
	int  result;
 
	while(size < BUFFER_SIZE)
	{
		result = ov_read(&oggStream, data + size, BUFFER_SIZE - size, 0, 2, 1, & section);
    
		if(result > 0)
			size += result;
		else
			if(result < 0)
				throw oggString(result);
		else
			break;
	}
    
	if(size == 0)
		return false;
 
	alBufferData(buffer, format, data, size, vorbisInfo->rate);
	check();*/
}

/**
 * Init and shutdown
 */
void al_mus_init()
{
	ALuint error;

	if(mus_inited)
		return;

	// We need a source that's never been used before, so allocate it here
	// This is a bit wasteful, but there's not much I can do about that...
	source_handle = al_src_alloc(SRCPRI_STREAM, -2, 0);
	if(source_handle == -1)
		return;

	// Lock the source so nobody else can use it, and get the raw source
	al_src_lock(source_handle);
	source = al_src_get(source_handle);

	// Set some source parameters
	alSource3f(source, AL_POSITION,        0.0, 0.0, 0.0);
	alSource3f(source, AL_VELOCITY,        0.0, 0.0, 0.0);
	alSource3f(source, AL_DIRECTION,       0.0, 0.0, 0.0);
	alSourcef (source, AL_ROLLOFF_FACTOR,  0.0          );
	alSourcei (source, AL_SOURCE_RELATIVE, AL_TRUE      );

	// And the Ogg Vorbis reader callbacks
	oggcallback_struct.read_func  = oggcallback_read;
	oggcallback_struct.close_func = oggcallback_close;
	oggcallback_struct.seek_func  = oggcallback_seek;
	oggcallback_struct.tell_func  = oggcallback_tell;

	mus_inited = qtrue;
}

void al_mus_shutdown()
{
	if(!mus_inited)
		return;

	// Release the output source
	al_src_unlock(source_handle);
	source = 0;
	source_handle = -1;

	mus_inited = qfalse;
}

/**
 * Background music playback
 */

char *findExtension(const char *fni)
{
	char *fn = (char *)fni;
	char *eptr = NULL;
	while(*fn)
	{
		if(*fn == '.')
			eptr = fn;
		fn++;
	}

	return eptr;
}

void S_StartBackgroundTrack( const char *intro, const char *loop )
{
	const char *fn;
	qboolean result = qfalse;

	if(!mus_inited)
		return;

	// Stop any existing music that might be playing
	S_StopBackgroundTrack();

	// Find a filename
	if((intro) && (intro[0]))
		fn = intro;
	else if((loop) && (loop[0]))
		fn = loop;
	else
		return;

	// Try to find a file extension
	char *ext = findExtension(fn);
	if(ext)
	{
		if(!Q_stricmp(ext, ".ogg"))
			result = StartVorbisTrack(intro, loop);
		else if (!Q_stricmp(ext, ".wav"))
			result = StartWaveTrack(intro, loop);
		else
			Com_Printf(S_COLOR_YELLOW "WARNING: Unknown music extension %s\n", ext);
	}
	else
	{
		result = StartWaveTrack(intro, loop);
	}

	// If it loaded, start playing
	if(result)
	{
		// Generate the buffers
		alGenBuffers(BUFFERS, buffers);
	
		// Queue the buffers up
		alSourceQueueBuffers(source, BUFFERS, buffers);
	
		// Start playing
		alSourcePlay(source);

		mus_playing = qtrue;
	}
}

void S_StopBackgroundTrack( void )
{
	if(!mus_inited)
		return;

	if(!mus_playing)
		return;

	// Stop playing
	alSourceStop(source);

	// De-queue the buffers
	alSourceUnqueueBuffers(source, BUFFERS, buffers);

	// Destroy the buffers
	alDeleteBuffers(BUFFERS, buffers);

	if(mus_vorbis)
	{
		KillVorbisTrack();
	}
	else
	{
		StopWaveTrack();
	}

	mus_playing = qfalse;
}

void S_UpdateBackgroundTrack( void )
{

}

void al_mus_update(void)
{
	int processed;
	ALint state;
	static float gain = 0.5f;

	if(!mus_playing)
		return;

	alGetSourcei(source, AL_BUFFERS_PROCESSED, &processed);
	if(processed)
	{
		while(processed--)
		{
			ALuint b;
	
			alSourceUnqueueBuffers(source, 1, &b);

			if(mus_vorbis)
			{
				UpdateVorbisTrack(b);
			}
			else
			{
				UpdateWaveTrack(b);
			}

			alSourceQueueBuffers(source, 1, &b);
		}
	}

	// If it's not still playing, give it a kick
	alGetSourcei(source, AL_SOURCE_STATE, &state);
	if(state == AL_STOPPED)
		alSourcePlay(source);

	// Then set the gain property
	gain = (gain + (s_musicVolume->value * 2))/4.0f;
	alSourcef(source, AL_GAIN, gain * 0.8);
}
