/*
===========================================================================
Copyright (C) 1999-2005 Id Software, Inc.
Copyright (C) 2005 Stuart Dalton (badcdev@gmail.com)

This file is part of Quake III Arena source code.

Quake III Arena source code is free software; you can redistribute it
and/or modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of the License,
or (at your option) any later version.

Quake III Arena source code is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Foobar; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
===========================================================================
*/

/*
 * snd_al_sources.c
 * Sources allocation and management for the OpenAL sound system
 *
 * 2005-08-23
 * Started again from scratch
 * Implemented priority-based channel allocation system
 * Fixed a lot of bugs
 * Added dynamic source creation (s_sources)
 */

#include "../client/client.h"
#include "snd_al_local.h"

typedef struct src_s src_t;
struct src_s
{
	ALuint source;		// OpenAL source object
	sfxHandle_t sfx;	// Sound effect in use

	int lastUse;		// Last time used
	int priority;		// Priority
	int entity;		// Owning entity (-1 if none)
	int channel;		// Associated channel (-1 if none)

	int isActive;		// Is this source currently in use?
	int isLocked;		// This is locked (un-allocatable)
	int isLooping;		// Is this a looping effect (attached to an entity)
	int isTracking;		// Is this object tracking it's owner
};

#define MAX_SRC 128
static src_t srclist[MAX_SRC];
static int src_count = 0;
static qboolean src_inited = qfalse;

static int ambient_count = 0;

typedef struct sentity_s
{
	vec3_t origin;		// Object position

	int has_sfx;		// Associated sound source
	int sfx;
	int touched;		// Sound present this update?
} sentity_t;
static sentity_t entlist[MAX_GENTITIES];

static void al_src_kill(srcHandle_t src);

/**
 * Startup and shutdown
 */
qboolean al_src_init()
{
	int i;
	ALenum error;

	if(src_inited)
		return qtrue;

	// Clear the sources data structure
	Com_Memset(srclist, 0, sizeof(srclist));

	// Cap s_sources to MAX_SRC
	if(s_sources->integer > MAX_SRC)
		Cvar_SetValue("s_sources", MAX_SRC);
	else if(s_sources->integer < 0)
		Cvar_SetValue("s_sources", 64);

	// Allocate as many sources as possible
	for(i = 0; i < s_sources->integer; i++)
	{
		alGenSources(1, &srclist[i].source);
		if((error = alGetError()) != AL_NO_ERROR)
			break;
		src_count++;
	}

	// All done. Print this for informational purposes
	Com_Printf("allocated %d sources\n", src_count);
	src_inited = qtrue;
}

void al_src_shutdown()
{
	int i;

	if(!src_inited)
		return;

	// Destroy all the sources
	for(i = 0; i < src_count; i++)
		al_src_kill(i);

	src_inited = qfalse;
}

/**
 * Source setup
 */
static void al_src_setup(srcHandle_t src, sfxHandle_t sfx, int priority, int entity, int channel, qboolean local)
{
	ALuint buffer;
	float null_vector[] = {0, 0, 0};

	// Mark the SFX as used, and grab the raw AL buffer
	al_buf_use(sfx);
	buffer = al_buf_get(sfx);

	// Set up src struct
	srclist[src].lastUse = Com_Milliseconds();
	srclist[src].sfx = sfx;
	srclist[src].priority = priority;
	srclist[src].entity = entity;
	srclist[src].channel = channel;
	srclist[src].isActive = qtrue;
	srclist[src].isLocked = qfalse;
	srclist[src].isLooping = qfalse;
	srclist[src].isTracking = qfalse;

	// Set up OpenAL source
	alSourcei(srclist[src].source, AL_BUFFER, buffer);
	alSourcef(srclist[src].source, AL_PITCH, 1.0f);
	alSourcef(srclist[src].source, AL_GAIN, s_gain->value * s_volume->value);
	alSourcefv(srclist[src].source, AL_POSITION, null_vector);
	alSourcefv(srclist[src].source, AL_VELOCITY, null_vector);
	alSourcei(srclist[src].source, AL_LOOPING, AL_FALSE);
	alSourcef(srclist[src].source, AL_REFERENCE_DISTANCE, s_minDistance->value);

	if(local)
	{
		alSourcei(srclist[src].source, AL_SOURCE_RELATIVE, AL_TRUE);
		alSourcef(srclist[src].source, AL_ROLLOFF_FACTOR, 0);
	}
	else
	{
		alSourcei(srclist[src].source, AL_SOURCE_RELATIVE, AL_FALSE);
		alSourcef(srclist[src].source, AL_ROLLOFF_FACTOR, s_rolloff->value);
	}
}

static void al_src_kill(srcHandle_t src)
{
	// I'm not touching it. Unlock it first.
	if(srclist[src].isLocked)
		return;

	// Stop it if it's playing
	if(srclist[src].isActive)
		alSourceStop(srclist[src].source);

	// Remove the entity association
	if((srclist[src].isLooping) && (srclist[src].entity != -1))
	{
		int ent = srclist[src].entity;
		entlist[ent].has_sfx = 0;
		entlist[ent].sfx = -1;
		entlist[ent].touched = qfalse;
	}

	srclist[src].sfx = 0;
	srclist[src].lastUse = 0;
	srclist[src].priority = 0;
	srclist[src].entity = -1;
	srclist[src].channel = -1;
	srclist[src].isActive = qfalse;
	srclist[src].isLocked = qfalse;
	srclist[src].isLooping = qfalse;
	srclist[src].isTracking = qfalse;
}

/**
 * Source allocation
 */
srcHandle_t al_src_alloc(int priority, int entnum, int channel)
{
	int i;
	int empty = -1;
	int weakest = -1;
	int weakest_time = Com_Milliseconds();
	int weakest_pri = 999;

	for(i = 0; i < src_count; i++)
	{
		// If it's locked, we aren't even going to look at it
		if(srclist[i].isLocked)
			continue;

		// Is it empty or not?
		if((!srclist[i].isActive) && (empty == -1))
			empty = i;
		else if(srclist[i].priority < priority)
		{
			// If it's older or has lower priority, flag it as weak
			if((srclist[i].priority < weakest_pri) ||
				(srclist[i].lastUse < weakest_time))
			{
				weakest_pri = srclist[i].priority;
				weakest_time = srclist[i].lastUse;
				weakest = i;
			}
		}

		// Is it an exact match, and not on channel 0?
		if((srclist[i].entity == entnum) && (srclist[i].channel == channel) && (channel != 0))
		{
			al_src_kill(i);
			return i;
		}
	}

	// Do we have an empty one?
	if(empty != -1)
	{
		return empty;
	}

	// No. How about an overridable one?
	if(weakest != -1)
	{
		al_src_kill(i);
		return weakest;
	}

	// Nothing. Return failure (cries...)
	return -1;
}

// Finds an active source with matching entity and channel numbers
// Returns -1 if there isn't one
srcHandle_t al_src_find(int entnum, int channel)
{
	int i;
	for(i = 0; i < src_count; i++)
	{
		if(!srclist[i].isActive)
			continue;
		if((srclist[i].entity == entnum) && (srclist[i].channel == channel))
			return i;
	}
	return -1;
}

// Locks and unlocks a source
// Locked sources will not be automatically reallocated or managed
// Once unlocked, the source may be reallocated again
void al_src_lock(srcHandle_t src)
{
	srclist[src].isLocked = qtrue;
}

void al_src_unlock(srcHandle_t src)
{
	srclist[src].isLocked = qfalse;
}

// Entity position management
void al_src_entity_move(int entnum, const vec3_t origin)
{
	if ( entnum < 0 || entnum > MAX_GENTITIES ) {
		Com_Error( ERR_DROP, "S_UpdateEntityPosition: bad entitynum %i", entnum );
	}
	VectorCopy( origin, entlist[entnum].origin );
}

// Play a local (non-spatialized) sound effect
void al_src_local(sfxHandle_t sfx, int channel)
{
	// Try to grab a source
	srcHandle_t src = al_src_alloc(SRCPRI_LOCAL, -1, channel);
	if(src == -1)
		return;

	// Set up the effect
	al_src_setup(src, sfx, SRCPRI_LOCAL, -1, channel, qtrue);

	// Start it playing
	alSourcePlay(srclist[src].source);
}

// Play a one-shot sound effect
void al_src_play(sfxHandle_t sfx, vec3_t origin, int entnum, int entchannel)
{
	// Try to grab a source
	srcHandle_t src = al_src_alloc(SRCPRI_ONESHOT, entnum, entchannel);
	if(src == -1)
		return;

	// Set up the effect
	al_src_setup(src, sfx, SRCPRI_ONESHOT, entnum, entchannel, qfalse);

	if(origin == NULL)
	{
		srclist[src].isTracking = qtrue;
		alSourcefv(srclist[src].source, AL_POSITION, entlist[entnum].origin);
	}
	else
	{
		alSourcefv(srclist[src].source, AL_POSITION, origin);
	}

	// Start it playing
	alSourcePlay(srclist[src].source);
}

// Start a looping sound effect
void al_src_loop_clear()
{
	int i;
	for(i = 0; i < src_count; i++)
	{
		if((srclist[i].isLooping) && (srclist[i].entity != -1))
			entlist[srclist[i].entity].touched = qfalse;
	}
}

void al_src_loop(int priority, sfxHandle_t sfx, const vec3_t origin, const vec3_t velocity, int entnum)
{
	int src;
	qboolean need_to_play = qfalse;
	float null_vector[] = {0.0f, 0.0f, 0.0f};

	// Do we need to start a new sound playing?
	if(!entlist[entnum].has_sfx)
	{
		// Try to get a channel
		ambient_count++;
		src = al_src_alloc(priority, entnum, -1);
		if(src == -1)
			return;
		need_to_play = qtrue;
	}
	else if(srclist[entlist[entnum].sfx].sfx != sfx)
	{
		// Need to restart. Just re-use this channel
		src = entlist[entnum].sfx;
		al_src_kill(src);
		need_to_play = qtrue;
	}
	else
		src = entlist[entnum].sfx;

	if(need_to_play)
	{
		// Set up the effect
		al_src_setup(src, sfx, priority, entnum, -1, qfalse);
		alSourcei(srclist[src].source, AL_LOOPING, AL_TRUE);
		srclist[src].isLooping = qtrue;

		// Set up the entity
		entlist[entnum].has_sfx = qtrue;
		entlist[entnum].sfx = src;
		need_to_play = qtrue;
	}

	// Set up the position and velocity
	alSourcefv(srclist[src].source, AL_POSITION, origin);
	if(s_doppler->integer == 1)
		alSourcefv(srclist[src].source, AL_VELOCITY, velocity);
	else
		alSourcefv(srclist[src].source, AL_VELOCITY, null_vector);

	// Flag it
	entlist[entnum].touched = qtrue;

	// Play if need be
	if(need_to_play)
		alSourcePlay(srclist[src].source);
}

void al_src_loop_stop(int entnum)
{
	if(entlist[entnum].has_sfx)
		al_src_kill(entlist[entnum].sfx);
}

// Update state (move things around, manage sources, and so on)
void al_src_update()
{
	int i;
	ALint state;

	for(i = 0; i < src_count; i++)
	{
		if(srclist[i].isActive)
		{
			// Check if it's done, and flag it
			alGetSourcei(srclist[i].source, AL_SOURCE_STATE, &state);
			if(state == AL_STOPPED)
			{
				srclist[i].isActive = qfalse;
				al_src_kill(i);
				continue;
			}

			int ent = srclist[i].entity;

			// If a looping effect hasn't been touched this frame, kill it
			if(srclist[i].isLooping)
			{
				if(!entlist[ent].touched)
				{
					ambient_count--;
					al_src_kill(i);
				}
				continue;
			}

			// See if it needs to be moved
			if(srclist[i].isTracking)
				alSourcefv(srclist[i].source, AL_POSITION, entlist[ent].origin);
		}
	}
}

void al_src_shutup()
{
	int i;
	for(i = 0; i < src_count; i++)
		al_src_kill(i);
}

ALuint al_src_get(srcHandle_t src)
{
	return srclist[src].source;
}
